/*
 * Copyright (C) 2007 Koen Deforche
 *
 * See the LICENSE file for terms of use.
 */

#include "SimpleChatWidget.h"
#include "SimpleChatServer.h"

#include <WApplication>
#include <WContainerWidget>
#include <WLineEdit>
#include <WTable>
#include <WTableCell>
#include <WText>
#include <WTextArea>
#include <WPushButton>

#include <iostream>

using namespace Wt;

SimpleChatWidget::SimpleChatWidget(SimpleChatServer& server,
				   Wt::WContainerWidget *parent)
  : WCompositeWidget(parent),
    server_(server),
    app_(WApplication::instance())
{
  setImplementation(layout_ = new WContainerWidget());

  setPositionScheme(Relative);

  user_ = server_.suggestGuest();
  letLogin();

  app_->enableUpdates();
}

SimpleChatWidget::~SimpleChatWidget()
{
  logout();
}

void SimpleChatWidget::letLogin()
{
  layout_->clear();

  new WText("User name: ", layout_);
  userNameEdit_ = new WLineEdit(user_, layout_);
  userNameEdit_->setFocus();

  WPushButton *b = new WPushButton("Login", layout_);
  b->setMargin(5, WWidget::Left);

  b->clicked.connect(SLOT(this, SimpleChatWidget::login));
  userNameEdit_->enterPressed.connect(SLOT(this, SimpleChatWidget::login));

  statusMsg_ = new WText(layout_);
  statusMsg_->setFormatting(WText::PlainFormatting);
  statusMsg_->setInline(false);
}

void SimpleChatWidget::login()
{
  WString name = WWebWidget::escapeText(userNameEdit_->text());

  if (!startChat(name))
    statusMsg_->setText("Sorry, name '" + name + "' is already taken.");
}

void SimpleChatWidget::logout()
{
  if (eventConnection_.connected()) {
    server_.logout(user_);
    eventConnection_.disconnect();

    messageEditArea_->clear();

    WPushButton *b = new WPushButton("Login again", messageEditArea_);
    b->clicked.connect(SLOT(this, SimpleChatWidget::letLogin));
  }
}

bool SimpleChatWidget::startChat(const WString& user)
{
  if (server_.login(user)) {
    eventConnection_
      = server_.chatEvent.connect(SLOT(this,
				       SimpleChatWidget::processChatEvent));
    user_ = user;    

    static const int MESSAGE_EDIT_AREA_HEIGHT = 80;
    static const int USERS_WIDTH_PERCENTAGE = 15;

    layout_->clear();

    messages_ = new WContainerWidget(layout_);
    messages_->setStyleClass("chat-msgs");
    messages_->setPositionScheme(Absolute);
    messages_->setOffsets(0, Left | Top);
    messages_
      ->resize(WLength(100 - USERS_WIDTH_PERCENTAGE - 1, WLength::Percentage),
	       height().value() - MESSAGE_EDIT_AREA_HEIGHT - 3);
    messages_->setOverflow(WContainerWidget::OverflowAuto, Vertical);

    messageEditArea_ = new WContainerWidget(layout_);
    messageEditArea_->setPositionScheme(Absolute);
    messageEditArea_->setOffsets(0, Left | Bottom);
    messageEditArea_->resize(WLength(100, WLength::Percentage),
			     MESSAGE_EDIT_AREA_HEIGHT);

    WContainerWidget *a = new WContainerWidget(messageEditArea_);
    a->setContentAlignment(AlignCenter);

    messageEdit_ = new WTextArea(a);
    messageEdit_->setStyleClass("chat-noedit");
    messageEdit_->setRows(2);
    messageEdit_->resize(WLength(99, WLength::Percentage), WLength());
    messageEdit_->setFocus();

    sendButton_ = new WPushButton("Send", messageEditArea_);

    sendButton_->clicked.connect(SLOT(sendButton_, WPushButton::disable));
    sendButton_->clicked.connect(SLOT(messageEdit_, WTextArea::disable));
    sendButton_->clicked.connect(SLOT(this, SimpleChatWidget::send));

    messageEdit_->enterPressed.connect(SLOT(sendButton_, WPushButton::disable));
    messageEdit_->enterPressed.connect(SLOT(messageEdit_, WTextArea::disable));
    messageEdit_->enterPressed.connect(SLOT(this, SimpleChatWidget::send));

    WPushButton *b = new WPushButton("Logout", messageEditArea_);
    b->clicked.connect(SLOT(this, SimpleChatWidget::logout));

    userList_ = new WContainerWidget(layout_);
    userList_->setStyleClass("chat-users");
    userList_->setPositionScheme(Absolute);
    userList_->setOffsets(0, Top | Right);
    userList_->resize(WLength(USERS_WIDTH_PERCENTAGE, WLength::Percentage),
		      height().value() - MESSAGE_EDIT_AREA_HEIGHT - 3);

    WText *w
      = new WText(false,
		  "<span class='chat-info'>You are joining the conversation as "
		  + user_ + "</span>", messages_);
    w->setStyleClass("chat-msg");

    updateUsers();
    
    return true;
  } else
    return false;
}

void SimpleChatWidget::send()
{
  if (!messageEdit_->text().empty()) {
    server_.sendMessage(user_, messageEdit_->text());
    messageEdit_->setText("");
  }

  messageEdit_->enable();
  messageEdit_->setFocus();
  sendButton_->enable();
}

void SimpleChatWidget::updateUsers()
{
  userList_->clear();

  SimpleChatServer::UserSet users = server_.users();

  WString usersStr;

  for (SimpleChatServer::UserSet::iterator i = users.begin();
       i != users.end(); ++i) {
    if (*i == user_)
      usersStr += "<span class='chat-self'>" + *i + "</span><br />";
    else
      usersStr += *i + "<br />";
  }

  userList_->addWidget(new WText(false, usersStr));
}

void SimpleChatWidget::processChatEvent(const ChatEvent& event)
{
  /*
   * This is where the "server-push" happens. This method is called
   * when a new event or message needs to be notified to the user. In
   * general, it is called from another session.
   *
   * First, we take the lock to safely manipulate the UI outside of the
   * normal event loop.
   */
  WApplication::UpdateLock lock = app_->getUpdateLock();

  WText *w = new WText(false, event.formattedHTML(user_), messages_);
  w->setStyleClass("chat-msg");

  /* no more than 100 messages back-log */
  if (messages_->count() > 100)
    delete messages_->children()[0];

  if (event.type() != ChatEvent::Message)
    updateUsers();

  /*
   * little javascript trick to make sure we scroll along with new content
   */
  app_->doJavaScript(messages_->jsRef() + ".scrollTop += "
		     + messages_->jsRef() + ".scrollHeight;");

  /*
   * Update the user-interface with these changes
   */
  app_->triggerUpdate();
}
